/*
 * stmt.c - Statement construction and execution
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "dwarfx.h"

#include "data.h"
#include "array.h"
#include "io.h"
#include "sprintf.h"
#include "stmt.h"


static const struct stmt *ip = NULL;


/* ----- Fix statement pointers -------------------------------------------- */


static void end_sequence(struct stmt **stmt,struct stmt *next)
{
    int more;

    if (!*stmt) {
	*stmt = next;
	return;
    }
    more = 1;
    while (more) {
	if (!(*stmt)->next) {
	    (*stmt)->next = next;
	    more = 0;
	}
	if ((*stmt)->fn == stmt_if) {
	    end_sequence(&(*stmt)->u.if_stmt.true,(*stmt)->next);
	    end_sequence(&(*stmt)->u.if_stmt.false,(*stmt)->next);
	}
	else if ((*stmt)->fn == stmt_while)
	    end_sequence(&(*stmt)->u.while_stmt.body,*stmt);
	else if ((*stmt)->fn == stmt_break)
	    (*stmt)->u.break_stmt.loop = (*stmt)->u.break_stmt.loop->next;
	stmt = &(*stmt)->next;
    }
}


void fix_statements(struct stmt **stmt)
{
    end_sequence(stmt,NULL);
}


/* ----- Statement execution ----------------------------------------------- */


void stmt_expression(const struct stmt *self)
{
    /* @@@ memory leak */
    evaluate_lvalue(self->u.expr.expr);
}


void stmt_print(const struct stmt *self)
{
    const struct without *wo;
    const struct arg_list *arg;

    for (wo = self->u.print.without; wo; wo = wo->next) {
	struct value value;

	value = evaluate(wo->expr);
	dwx_output_skip(value.value,dwx_sizeof(value.type));
    }
    for (arg = self->u.print.args; arg; arg = arg->next) {
	struct value value;

	value = evaluate(arg->expr);
	if (value.type == &string_type)
	    printf("%s\n",(const char *) value.value);
	else dwx_output_data(stdout,value.type,value.value);
    }
    dwx_output_all();
}


void stmt_printf(const struct stmt *self)
{
    const char *s;

    s = common_sprintf(self->u.printf.fmt,self->u.printf.args);
    printf("%s\n",s);
    free((char *) s);
}


#if 0
void stmt_read(const struct stmt *self)
{
    struct value value;
    const char *s;
    int n;
    const void *p;

    value = evaluate(self->u.read.expr);
    if (value.type == &type_type) {
	in_skip(dwx_sizeof(value_type(value)));
	return;
    }
    s = value_string(value);
    n = strlen(s)+1;
    p = in_read(n);
    if (strcmp(s,p)) {
	fprintf(stderr,"read mismatch\n");
	exit(1);
    }
    in_free(p);
}
#endif


void stmt_if(const struct stmt *self)
{
    ip = int_eval(self->u.if_stmt.expr) ?
      self->u.if_stmt.true : self->u.if_stmt.false;
}


void stmt_break(const struct stmt *self)
{
    ip = self->u.break_stmt.loop;
}


void stmt_exit(const struct stmt *self)
{
    if (!in_eof()) fprintf(stderr,"warning: unread dump data\n");
    exit(self->u.exit.expr ? int_eval(self->u.exit.expr) : 0);
}


void stmt_die(const struct stmt *self)
{
    if (self->u.die.expr) fprintf(stderr,"%s\n",string_eval(self->u.die.expr));
    else fprintf(stderr,"Died at line %u\n",self->u.die.line);
    exit(1);
}


void stmt_while(const struct stmt *self)
{
    if (int_eval(self->u.while_stmt.expr)) ip = self->u.while_stmt.body;
}


void execute(const struct stmt *pgm)
{
    ip = pgm;
    while (ip) {
	const struct stmt *stmt = ip;

//fprintf(stderr,"%p (%p)\n",ip,ip->fn);
	ip = ip->next;
	stmt->fn(stmt);
    }
}
